from enum import Enum
from typing import Any

from rich.console import Console
from typer import Exit

__all__ = [
    "ConsoleError",
    "Verbosity",
    "console_print",
    "set_verbosity",
]

console = Console(legacy_windows=False)


class Verbosity(Enum):
    quiet = 0
    normal = 1
    verbose = 2


_internal_verbosity = Verbosity.normal


def set_verbosity(mode: Verbosity) -> None:
    """Globally set the verbosity level for console printing."""
    global _internal_verbosity  # noqa: PLW0603
    _internal_verbosity = mode


def console_print(
    *objects: Any,
    level: Verbosity = Verbosity.normal,
    **kwargs: Any,
) -> None:
    """Print to console if the current verbosity level is sufficient."""
    if _internal_verbosity.value >= level.value:
        console.print(*objects, **kwargs)


class ConsoleError(Exit):
    def __init__(self, *objects: Any, **kwargs: Any) -> None:
        """Raise a user-friendly console error and print the message to the console."""
        console.print(*objects, **kwargs)
        super().__init__(code=1)



import sys
from pathlib import Path

from packaging.utils import parse_wheel_filename

from uv_pack._process import exit_on_error, run_cmd


def build_src_wheel(*, source_path: Path, out_path: Path, other_args: str) -> None:
    cmd = [
        "uv",
        "build",
        str(source_path),
        "--wheel",
        "--out-dir",
        str(out_path),
    ]

    cmd.extend(other_args.split())
    exit_on_error(run_cmd(cmd, "uv build"))


def build_requirements(
    *,
    requirements_txt: Path,
    requirements_export_txt: Path,
    vendor_directory: Path,
) -> None:
    vendor_pins: list[str] = []

    for wheel in vendor_directory.glob("*.whl"):
        name, version, *_ = parse_wheel_filename(wheel.name)
        vendor_pins.append(f"{name}=={version}")

    header = "\n".join(
        [
            "# This file was autogenerated by uv-pack via the following command:",
            f"#    uv-pack {' '.join(sys.argv[1:])}",
            *sorted(set(vendor_pins)),
        ],
    )

    requirements_txt.write_text(
        header + "\n" + requirements_export_txt.read_text(encoding="utf-8"),
        encoding="utf-8",
    )



from dataclasses import dataclass
from pathlib import Path

__all__ = [
    "PackLayout",
]


@dataclass(frozen=True, slots=True)
class PackLayout:
    """Layout of the output directory for a uv-pack."""

    output_directory: Path
    gitignore_file: Path
    requirements_txt: Path
    requirements_export_txt: Path
    requirements_local_txt: Path
    _wheels_dir: Path
    _vendor_dir: Path
    _python_dir: Path

    @property
    def wheels_dir(self) -> Path:
        self._wheels_dir.mkdir(exist_ok=True)
        return self._wheels_dir

    @property
    def vendor_dir(self) -> Path:
        self._vendor_dir.mkdir(exist_ok=True)
        return self._vendor_dir

    @property
    def python_dir(self) -> Path:
        self._python_dir.mkdir(exist_ok=True)
        return self._python_dir

    @classmethod
    def create(cls, output_directory: Path) -> "PackLayout":
        output_directory.mkdir(parents=True, exist_ok=True)
        wheels_dir = output_directory / "wheels"
        vendor_dir = output_directory / "vendor"
        python_dir = output_directory / "python"

        gitignore_file = output_directory / ".gitignore"
        gitignore_file.write_text("*", encoding="utf-8")

        return cls(
            output_directory,
            gitignore_file=gitignore_file,
            requirements_txt=output_directory / "requirements.txt",
            requirements_export_txt=wheels_dir / "requirements.txt",
            requirements_local_txt=vendor_dir / "requirements.txt",
            _wheels_dir=wheels_dir,
            _vendor_dir=vendor_dir,
            _python_dir=python_dir,
        )



from pathlib import Path

from packaging.utils import parse_wheel_filename

from uv_pack._process import exit_on_error, run_cmd


def download_third_party_wheels(
    *,
    requirements_file: Path,
    wheels_directory: Path,
    other_args: str,
) -> None:
    wheels_directory.mkdir(parents=True, exist_ok=True)

    cmd = [
        "uv",
        "run",
        "--with",
        "pip",
        "python",
        "-m",
        "pip",
        "download",
        "--prefer-binary",
        "--no-deps",
        "--disable-pip-version-check",
        "-r",
        str(requirements_file),
        "-d",
        str(wheels_directory),
    ]

    cmd.extend(other_args.split())
    exit_on_error(run_cmd(cmd, "pip download"))


def determine_download_requirements(
    *,
    requirements_file: Path,
    wheels_directory: Path,
) -> None:
    available_wheels: list[str] = []

    for wheel in wheels_directory.glob("*.whl"):
        name, version, *_ = parse_wheel_filename(wheel.name)
        available_wheels.append(f"{name}=={version}")

    requirements_file.write_text(
        "\n".join(sorted(set(available_wheels))),
        encoding="utf-8",
    )



from pathlib import Path

from uv_pack._process import exit_on_error, run_cmd


def export_requirements(
    *,
    requirements_file: Path,
    other_args: str,
) -> None:
    cmd = [
        "uv",
        "export",
        "--quiet",
        "--no-dev",
        "--no-header",
        "--no-hashes",
        "--no-emit-local",
        "--format=requirements.txt",
        f"--output-file={requirements_file}",
    ]

    cmd.extend(other_args.split())
    exit_on_error(run_cmd(cmd, "uv export"))


def export_local_requirements(
    *,
    requirements_file: Path,
    other_args: str,
) -> None:
    """Export only local packages to a plain requirements.txt file (each line is a local requirement path)."""
    cmd = [
        "uv",
        "export",
        "--quiet",
        "--no-dev",
        "--no-header",
        "--no-hashes",
        "--no-annotate",
        "--no-editable",
        "--only-emit-local",
        "--format=requirements.txt",
        f"--output-file={requirements_file}",
    ]

    cmd.extend(other_args.split())
    exit_on_error(run_cmd(cmd, "uv export"))



"""uv-pack: Bundle a locked uv environment into an offline-installable bundle.

Pipeline:
1. Clean the output directory
2. Export locked requirements via uv
3. Download third-party wheels into ./wheels
4. Build local workspace packages into ./vendor
5. Download a python interpreter to ./python

Result:
pack/
├── requirements.txt
├── wheels/              # third-party wheels
│   └── requirements.txt # index packages
├── vendor/              # locally built wheels
│   └── requirements.txt # local packages
├── python/              # python interpreter
├── unpack.sh
├── unpack.ps1
├── unpack.bat
├── .gitignore
└── README.md
"""

import shutil
from collections.abc import Iterable
from enum import Enum
from pathlib import Path

import typer

from uv_pack._build import build_requirements, build_src_wheel
from uv_pack._download import download_third_party_wheels
from uv_pack._export import export_local_requirements, export_requirements
from uv_pack._files import PackLayout
from uv_pack._logging import ConsoleError, Verbosity, console_print, set_verbosity
from uv_pack._process import run_step
from uv_pack._python import download_latest_python_build
from uv_pack._scripts import copy_unpack_scripts

# -----------------------------------------------------------------------------
# CLI setup
# -----------------------------------------------------------------------------

app: typer.Typer = typer.Typer(add_completion=True)


def main() -> None:
    """Main entry point for uv-pack CLI."""
    try:
        app()
    except ConsoleError:
        raise  # already formatted for user
    except KeyboardInterrupt as err:
        console_print("[yellow]Interrupted by user[/yellow]")
        raise typer.Exit(130) from err
    except Exception as err:
        msg = (
            "[bold red]✘ An unexpected internal error occurred, please try again or"
            " open an issue https://github.com/davnn/uv-pack/issues.[/bold red]"
        )
        raise ConsoleError(msg) from err


# -----------------------------------------------------------------------------
# Step model
# -----------------------------------------------------------------------------


class Step(str, Enum):
    """Steps to be performed in the pack pipeline."""

    clean = "clean"
    export = "export"
    download = "download"
    build = "build"
    python = "python"


PIPELINE_ORDER: tuple[Step, ...] = (
    Step.clean,
    Step.export,
    Step.download,
    Step.build,
    Step.python,
)


# -----------------------------------------------------------------------------
# Helper operations
# -----------------------------------------------------------------------------


def _additional_cli_args(cmd_name: str) -> str:
    return f"Additional command line arguments to be provided to '{cmd_name}'"


def _normalize_steps(
    steps: Iterable[Step] | None,
    skip: Iterable[Step] | None,
) -> list[Step]:
    selected = set(PIPELINE_ORDER) if steps is None else set(steps)
    selected = selected.difference([] if skip is None else set(skip))
    return [step for step in PIPELINE_ORDER if step in selected]


def _raise_requirement_txt_missing(path: Path) -> None:
    if not path.exists():
        msg = f"[bold red]✘ No requirements file found:[/bold red] '{path}', did you skip the 'export' step?"
        raise ConsoleError(msg)


# -----------------------------------------------------------------------------
# Orchestration command
# -----------------------------------------------------------------------------


@app.command()
def pack(
    *,
    steps: list[Step] | None = typer.Argument(
        PIPELINE_ORDER,
        help="Pipeline steps to run (multiple can be whitespace-separated)",
    ),
    skip: list[Step] | None = typer.Option(
        None,
        "--skip",
        "-s",
        help="Pipeline steps to skip (can be supplied multiple times)",
    ),
    output_directory: Path = typer.Option(
        Path("./pack"),
        "--output-directory",
        "-o",
        help="Path to output directory",
    ),
    uv_export: str = typer.Option(
        default="",
        help=_additional_cli_args("uv export"),
    ),
    pip_download: str = typer.Option(
        default="",
        help=_additional_cli_args("pip download"),
    ),
    uv_build: str = typer.Option(
        default="",
        help=_additional_cli_args("uv build"),
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Enable verbose output",
    ),
) -> None:
    """Pack a locked uv environment into an offline-installable bundle."""
    set_verbosity(Verbosity.verbose if verbose else Verbosity.normal)
    selected_steps = _normalize_steps(steps, skip)
    console_print(
        f"[dim]Running steps:[/dim] {[step.value for step in selected_steps]}",
    )

    if Step.clean in selected_steps:
        with run_step("clean"):
            shutil.rmtree(output_directory, ignore_errors=True)

        console_print(
            f"[green]✔ Cleaned[/green] output directory '{output_directory}'",
            level=Verbosity.verbose,
        )

    # initialize pack directory structure and copy unpack scripts
    pack = PackLayout.create(output_directory=output_directory)
    copy_unpack_scripts(output_directory=output_directory)

    if Step.export in selected_steps:
        with run_step("export"):
            export_requirements(
                requirements_file=pack.requirements_export_txt,
                other_args=uv_export,
            )
            export_local_requirements(
                requirements_file=pack.requirements_local_txt,
                other_args=uv_export,
            )
        console_print(
            f"[green]✔ Exported[/green] requirements '{pack.requirements_export_txt}'",
            level=Verbosity.verbose,
        )
        console_print(
            f"[green]✔ Exported[/green] requirements '{pack.requirements_local_txt}'",
            level=Verbosity.verbose,
        )

    if Step.download in selected_steps:
        _raise_requirement_txt_missing(pack.requirements_export_txt)

        with run_step("download"):
            download_third_party_wheels(
                requirements_file=pack.requirements_export_txt,
                wheels_directory=pack.wheels_dir,
                other_args=pip_download,
            )

    if Step.build in selected_steps:
        _raise_requirement_txt_missing(pack.requirements_local_txt)
        _raise_requirement_txt_missing(pack.requirements_export_txt)

        # show the progress for each package in verbose mode, otherwise a single progress report is shown
        with run_step("build", should_run=not verbose):
            for line in pack.requirements_local_txt.read_text(
                encoding="utf-8",
            ).splitlines():
                with run_step("build", should_run=verbose):
                    build_src_wheel(
                        source_path=Path(line),
                        out_path=pack.vendor_dir,
                        other_args=uv_build,
                    )
                console_print(
                    f"[green]✔ Built[/green] wheel: '{line}'",
                    level=Verbosity.verbose,
                )

            for sdist in pack.wheels_dir.glob("*.tar.gz"):
                with run_step("build", should_run=verbose):
                    build_src_wheel(
                        source_path=sdist,
                        out_path=pack.wheels_dir,
                        other_args=uv_build,
                    )
                    sdist.unlink(missing_ok=True)
                console_print(
                    f"[green]✔ Built[/green] wheel: '{sdist}'",
                    level=Verbosity.verbose,
                )

            build_requirements(
                requirements_txt=pack.requirements_txt,
                requirements_export_txt=pack.requirements_export_txt,
                vendor_directory=pack.vendor_dir,
            )
            console_print(
                f"[green]✔ Built[/green] requirements: '{pack.requirements_txt}'",
                level=Verbosity.verbose,
            )

    if Step.python in selected_steps:
        pack.python_dir.mkdir(exist_ok=True)
        python_path = download_latest_python_build(
            dest_dir=pack.python_dir,
        )
        console_print(
            f"[green]✔ Python[/green] archive: '{python_path}'",
            level=Verbosity.verbose,
        )

    console_print("[green]✔ Done[/green]")



import subprocess
import sys
from collections.abc import Iterator
from contextlib import contextmanager, nullcontext
from dataclasses import dataclass

import typer
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, TextColumn

from uv_pack._logging import console_print


@dataclass(frozen=True)
class CommandOutput:
    """Captured subprocess execution result."""

    name: str
    stdout: str
    stderr: str
    returncode: int


@contextmanager
def run_step(cmd_name: str, *, should_run: bool = True) -> Iterator[Progress | None]:
    """Context manager to show a progress display during step execution."""
    ctx = (
        Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
        )
        if should_run
        else nullcontext()
    )

    with ctx as progress:
        if progress is not None:
            progress.add_task(f"Running '{cmd_name}'...", total=None)
        yield progress


def run_cmd(cmd: list[str], cmd_name: str) -> CommandOutput:
    """Execute a subprocess command and capture output."""
    proc = subprocess.run(
        cmd,
        text=True,
        capture_output=True,
        check=False,
    )
    return CommandOutput(
        name=cmd_name,
        stdout=proc.stdout,
        stderr=proc.stderr,
        returncode=proc.returncode,
    )


def exit_on_error(result: CommandOutput) -> None:
    """Exit the CLI if a subprocess failed."""
    if result.returncode == 0:
        return

    if result.stderr:
        if sys.stderr.isatty():
            console_print(
                Panel.fit(
                    result.stderr.rstrip(),
                    title=f"[bold red]✘ '{result.name}' failed[/bold red]",
                    border_style="red",
                ),
            )
        else:
            typer.echo(result.stderr, err=True)

    raise typer.Exit(code=result.returncode)



"""Copy unpack scripts to create (or reuse) a virtual environment and install dependencies from the pack directory."""

import shutil
import stat
from importlib.resources import files
from pathlib import Path

__all__ = [
    "copy_unpack_scripts",
]


def copy_unpack_scripts(
    *,
    output_directory: Path,
) -> None:
    """Write unpack scripts into the pack directory."""
    scripts_dir = files("uv_pack") / "scripts"
    copy_file = lambda src: shutil.copyfile(str(src), str(output_directory / src.name))  # type: ignore

    output_directory.mkdir(parents=True, exist_ok=True)
    copy_file(scripts_dir / "unpack.sh")
    copy_file(scripts_dir / "unpack.ps1")
    copy_file(scripts_dir / "unpack.cmd")
    copy_file(scripts_dir / "README.md")
    _make_executable(output_directory / "unpack.sh")


def _make_executable(path: Path) -> None:
    """Best-effort make a script executable (POSIX)."""
    try:
        mode = path.stat().st_mode
        path.chmod(mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
    except OSError:
        pass
